import json
from datetime import datetime, timedelta
import random


# Функция для загрузки данных из JSON-файла и преобразования в граф
def load_graph_from_json(file_path):
    with open(file_path, 'r', encoding='utf-8') as file:
        data = json.load(file)

    graph = {}
    start_point = data['from']
    end_point = data['to']

    for flight in data['flights']:
        departure_point = flight['departurePoint']
        destination_point = flight['destinationPoint']
        distance = flight['distance']
        departure_time = datetime.fromisoformat(flight['departureTime'])
        destination_time = datetime.fromisoformat(flight['destinationTime'])

        if departure_point not in graph:
            graph[departure_point] = {}

        graph[departure_point][destination_point] = (distance, departure_time, destination_time)

    return graph, start_point, end_point


# Функция для вычисления длины и времени пути с учетом минимального интервала времени
def path_length_and_time(path, graph):
    length = 0
    start_time = graph[path[0]][path[1]][1]
    end_time = start_time
    for i in range(len(path) - 1):
        if path[i] not in graph or path[i + 1] not in graph[path[i]]:
            return float('inf'), start_time, end_time
        length += graph[path[i]][path[i + 1]][0]
        if i > 0:  # Пропуск первой вершины
            if end_time + timedelta(minutes=5) > graph[path[i]][path[i + 1]][1]:
                return float('inf'), start_time, end_time  # Неприемлемый путь
        end_time = graph[path[i]][path[i + 1]][2]
    return length, start_time, end_time


# Функция для генерации начальной популяции
def generate_population(size, start, end, graph):
    population = []
    for _ in range(size):
        path = [start]
        while path[-1] != end:
            if path[-1] not in graph or not graph[path[-1]]:
                break
            next_node = random.choice(list(graph[path[-1]].keys()))
            if next_node not in path:
                path.append(next_node)
        if path[-1] == end:
            population.append(path)
    return population


# Функция для селекции родителей
def select_parents(population, graph):
    sorted_population = sorted(population, key=lambda p: path_length_and_time(p, graph)[0])
    return sorted_population[:len(sorted_population) // 2]


# Функция для скрещивания (кроссинговера)
def crossover(parent1, parent2):
    crossover_point = len(parent1) // 2
    child = parent1[:crossover_point]
    for gene in parent2:
        if gene not in child:
            child.append(gene)
    return child


# Функция для мутации
def mutate(child):
    if len(child) <= 2:
        return child  # Если путь состоит из одной вершины, мутация не требуется
    index1 = random.randint(1, len(child) - 2)
    index2 = random.randint(1, len(child) - 2)
    child[index1], child[index2] = child[index2], child[index1]
    return child


# Главная функция генетического алгоритма
def genetic_algorithm(start, end, graph, population_size=100, generations=100):
    population = generate_population(population_size, start, end, graph)
    for _ in range(generations):
        parents = select_parents(population, graph)
        new_population = parents[:]
        while len(new_population) < population_size:
            parent1 = random.choice(parents)
            parent2 = random.choice(parents)
            child = crossover(parent1, parent2)
            if random.random() < 0.1:  # Вероятность мутации
                child = mutate(child)
            new_population.append(child)
        population = new_population
    best_path = min(population, key=lambda p: path_length_and_time(p, graph)[0])
    best_length, start_time, end_time = path_length_and_time(best_path, graph)
    return best_path, best_length, start_time, end_time


# Пример использования
file_path = './/data_ga.json'
graph, start_point, end_point = load_graph_from_json(file_path)

best_path, length, start_time, end_time = genetic_algorithm(start_point, end_point, graph)
print("Наиболее короткий путь от", start_point, "до", end_point, ":", best_path)
print("Длина пути:", length)
print("Начальное время:", start_time.strftime("%d %B %Y %H:%M"))
print("Конечное время:", end_time.strftime("%d %B %Y %H:%M"))
